#include "frmDirectWindow.h"

#include <OS.h>
#include <Application.h>
#include <GraphicsDefs.h>
#include <Screen.h>

#include <new>

#include <malloc.h>
#include <memory.h>

#include "memblt.h"
#include "helper.h"

// window ///////////////////////////////////////////////////////////////////////

FrmDirectWindow::FrmDirectWindow(FrameworkParams &params, status_t &error, BString &errorString) : 

	BDirectWindow(BRect(B_ORIGIN, params.frameBounds), "BDirectWindow Framework", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE),

	mFlipThreadID(B_ERROR),
	mFlipPort(B_ERROR),
	mBitsArea(B_ERROR),
	mBack(NULL)
{
	if (B_OK != error)
		return;

	mConnected = false;
	mNumRingBufs = params.numRingBuffers;
	mDrawSelector = 0;

	mFrameWidth = (int32) params.frameBounds.x + 1;
	mFrameHeight = (int32) params.frameBounds.y + 1;

	size_t pixel_chunk;
	size_t pixels_per_chunk;
	size_t row_alignment;

	::get_pixel_size_for(params.pixelFormat, &pixel_chunk, &row_alignment, &pixels_per_chunk);

	mBackBitsInfo.bytes_per_row = (mFrameWidth * pixel_chunk) / pixels_per_chunk;
	mBackBitsInfo.bits_per_pixel = (pixel_chunk * 8) / pixels_per_chunk;
	mBackBitsInfo.pixel_format = params.pixelFormat;

	void *area_bits;
	mBitsArea = ::kreate_area("frame_buffer",
							  &area_bits,
							  B_ANY_ADDRESS,
							  (mNumRingBufs * mFrameHeight * mBackBitsInfo.bytes_per_row + B_PAGE_SIZE - 1) & ((uint32) -1 ^ (B_PAGE_SIZE - 1)),
							  B_NO_LOCK,
							  B_READ_AREA | B_WRITE_AREA,
							  error, errorString);
	if (B_OK > mBitsArea)
		return;

	mBack = new(std::nothrow) BUFFER_RECORD [mNumRingBufs];
	if (NULL == mBack)
	{
		errorString = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	for (int32 i = 0; i < mNumRingBufs; i++)
	{
		mBack[i].writeLock = ::kreate_sem(0, NULL, error, errorString);
		mBack[i].bits = (uint8 *) area_bits + i * mFrameHeight * mBackBitsInfo.bytes_per_row;
		mBack[i].unlockAtConnect = true;
	}

	if (B_OK != error) // account for last ::kreate_sem()
		return;

	if (SupportsWindowMode())
	{
		BScreen *screen = new(std::nothrow) BScreen(this);
		if (NULL == screen)
		{
			errorString = "'bad_alloc' at framework construction.";
			error = B_ERROR;
			return;
		}

		BRect scrRect = screen->Frame();
		MoveTo((scrRect.Width() - params.frameBounds.x) * .5f, (scrRect.Height() - params.frameBounds.y) * .5f);
		delete screen;
	}
	else
		SetFullScreen(true);

	mFlipPort = ::kreate_port(mNumRingBufs, "flip_port", error, errorString);

	if (B_OK > mFlipPort)
		return;

	mFlipThreadID = ::spawn_thread(FlippingThread, "flipping_thread", B_DISPLAY_PRIORITY, (void *) this);

	if (B_OK > mFlipThreadID)
		switch(error = mFlipThreadID)
		{
		case B_NO_MORE_THREADS:

			errorString = "failed to spawn a thread: all thread identifiers are currently taken.";
			return;

		case B_NO_MEMORY:

			errorString = "failed to spawn a thread: not enough memory for the operation.";
			return;

		default:

			errorString = "failed to spawn a thread: alien error code.";
			return;
		}

	::resume_thread(mFlipThreadID);
}

FrmDirectWindow::~FrmDirectWindow()
{
	if (NULL != mBack)
		for (int i = 0; i < mNumRingBufs; i++)
			::delete_sem(mBack[i].writeLock);

	::delete_port(mFlipPort);

	int32 result; 
	::wait_for_thread(mFlipThreadID, &result);

	thread_id painter = ::find_thread(Framework::__painter_name());

	if (B_NAME_NOT_FOUND != painter)
		::wait_for_thread(painter, &result);

	::delete_area(mBitsArea);
	delete [] mBack;
}

bool FrmDirectWindow::QuitRequested()
{
	Hide(); 
	Sync();
	
 	::be_app_messenger.SendMessage(B_QUIT_REQUESTED);

	return BDirectWindow::QuitRequested();
}

void FrmDirectWindow::DirectConnected(direct_buffer_info *info)
{
	mFlipLock.Lock();

	switch((*info).buffer_state & B_DIRECT_MODE_MASK)
	{
		case B_DIRECT_START: 

			mFrontBits = (uint8 *) (*info).bits;

			mFrontBitsInfo.bytes_per_row  = (*info).bytes_per_row;
			mFrontBitsInfo.bits_per_pixel = (*info).bits_per_pixel;
			mFrontBitsInfo.pixel_format   = (*info).pixel_format;

			mFrontClipInfo.window_bounds = (*info).window_bounds;
			mFrontClipInfo.pclip_list = NULL;

			for (int i = 0; i < mNumRingBufs; i++)
				if (mBack[i].unlockAtConnect)
				{
					::release_sem_etc(mBack[i].writeLock, 1, B_DO_NOT_RESCHEDULE);
					mBack[i].unlockAtConnect = false;
				}

			mConnected = true;

		case B_DIRECT_MODIFY: 

			mFrontClipInfo.clip_list_count = (*info).clip_list_count;
			mFrontClipInfo.pclip_list =
				(clipping_rect *) ::realloc(mFrontClipInfo.pclip_list, mFrontClipInfo.clip_list_count * sizeof(clipping_rect));
			::memcpy(mFrontClipInfo.pclip_list, (*info).clip_list, mFrontClipInfo.clip_list_count * sizeof(clipping_rect));

			break;

		case B_DIRECT_STOP:

			::free(mFrontClipInfo.pclip_list);

			mConnected = false;
			break; 
	}

	mFlipLock.Unlock();
}

int32 FrmDirectWindow::FlippingThread(void *data)
{
	FrmDirectWindow *wnd = (FrmDirectWindow *) data; 
	int FlipSelector = 0;

	int32 command;

	while (B_BAD_PORT_ID != ::read_port((*wnd).mFlipPort, &command, NULL, 0))
	{
		(*wnd).mFlipLock.Lock();

		if ((*wnd).mConnected)
		{
			uint32 bytes_per_pixel = (*wnd).mFrontBitsInfo.bits_per_pixel >> 3;

			for (uint32 i = 0; i < (*wnd).mFrontClipInfo.clip_list_count; i++)
			{
				clipping_rect *clip = (*wnd).mFrontClipInfo.pclip_list + i;

				int32 clip_width = (*clip).right - (*clip).left + 1;
				int32 clip_height = (*clip).bottom - (*clip).top + 1;

				uint8 *dst = (*wnd).mFrontBits +
								(*clip).top * (*wnd).mFrontBitsInfo.bytes_per_row +
								(*clip).left * bytes_per_pixel;

				uint8 *src = (*wnd).mBack[FlipSelector].bits +
								((*clip).top - (*wnd).mFrontClipInfo.window_bounds.top) * (*wnd).mBackBitsInfo.bytes_per_row +
								((*clip).left - (*wnd).mFrontClipInfo.window_bounds.left) * bytes_per_pixel;

				for (int y = 0; y < clip_height; y++)
				{ 
//					::x86_memblt_long_unaligned(src, dst, clip_width * bytes_per_pixel);
					/* above appropriate for Write-Combine-enabled systems (P6-based, K6) */

//					::x86_memblt_long_aligned(src, dst, clip_width * bytes_per_pixel);
					/* above appropriate for Write-Combine-disabled systems (P5, P5MMX, K5) */					

					::any_memblt_32bytes_unrolled(src, dst, clip_width * bytes_per_pixel);
					/* above most appropriate for PII systems */

					dst += (*wnd).mFrontBitsInfo.bytes_per_row;
					src += (*wnd).mBackBitsInfo.bytes_per_row; 
				}
			}

			::release_sem((*wnd).mBack[FlipSelector].writeLock);
		}
		else
			(*wnd).mBack[FlipSelector].unlockAtConnect = true;

		(*wnd).mFlipLock.Unlock();

		++FlipSelector %= (*wnd).mNumRingBufs; 
	}

	return B_OK;
}

status_t FrmDirectWindow::frame_open(BUFFER_DESC &desc)
{
	status_t res = ::acquire_sem(mBack[mDrawSelector].writeLock);

	if (B_OK == res)
	{
		desc.frameHeight	= mFrameHeight;
		desc.frameWidth		= mFrameWidth;
		desc.pixelFormat	= mBackBitsInfo.pixel_format;
		desc.rowStride		= mBackBitsInfo.bytes_per_row;
		desc.pBits			= mBack[mDrawSelector].bits;
	}

	return res;
}

void FrmDirectWindow::frame_close()
{
	::write_port(mFlipPort, 0, NULL, 0);

	++mDrawSelector %= mNumRingBufs; 
}
