#include "frmWindowScreen.h"

#include <Application.h>
#include <memory.h>

#include <new>

#include "helper.h"

#define SANE_FLIP_PERIOD		18000	// usec, ~55.5Hz

// window ///////////////////////////////////////////////////////////////////////

uint32 FrmWindowScreen::MatchScreenSpace(BPoint &frameBounds, color_space color)
{
	static const int32 nSupportedResolutions = 6;
	static const int32 nSupportedBitDepths = 4;

	static const struct ScreenSpaceMapper
	{
		int32 width;
		int32 height;
		uint32 screen_space[nSupportedBitDepths];
	}
	mapper[nSupportedResolutions] = // keep in order with resolutions priority 
	{
		{  640,  480, { B_8_BIT_640x480,	B_15_BIT_640x480,	B_16_BIT_640x480,	B_32_BIT_640x480 }},
		{  800,  600, { B_8_BIT_800x600,	B_15_BIT_800x600,	B_16_BIT_800x600,	B_32_BIT_800x600 }},
		{ 1024,  768, { B_8_BIT_1024x768,	B_15_BIT_1024x768,	B_16_BIT_1024x768,	B_32_BIT_1024x768 }},
		{ 1152,  900, { B_8_BIT_1152x900,	B_15_BIT_1152x900,	B_16_BIT_1152x900,	B_32_BIT_1152x900 }},
		{ 1280, 1024, { B_8_BIT_1280x1024,	B_15_BIT_1280x1024,	B_16_BIT_1280x1024,	B_32_BIT_1280x1024 }},
		{ 1600, 1200, { B_8_BIT_1600x1200,	B_15_BIT_1600x1200,	B_16_BIT_1600x1200,	B_32_BIT_1600x1200 }}
	}; 

	for (int32 i = 0; i < nSupportedResolutions; i++)
	{
		if (((int32) frameBounds.x) < mapper[i].width &&
			((int32) frameBounds.y) < mapper[i].height)
		{
			switch(color)
			{
			case B_CMAP8:
			case B_GRAY8:
				return mapper[i].screen_space[0];
			case B_RGB15:
			case B_RGBA15:
			case B_RGB15_BIG:
			case B_RGBA15_BIG:
				return mapper[i].screen_space[1];
			case B_RGB16:
			case B_RGB16_BIG:
				return mapper[i].screen_space[2];
			case B_RGB32:
			case B_RGBA32:
			case B_RGB32_BIG:
			case B_RGBA32_BIG:
				return mapper[i].screen_space[3];
			default:
				return 0;
			}
		}
	}
	return 0;
}

FrmWindowScreen::FrmWindowScreen(FrameworkParams &params, status_t &error, BString &errorString) : 

	BWindowScreen("BWindowScreen Framework", MatchScreenSpace(params.frameBounds, params.pixelFormat), &error),

	mBScreen(NULL),
	mBits(NULL),
	mParams(params),
	mError(error),
	mErrorString(errorString)
{
	if (B_OK != error)
	{
		errorString = "failure constructing a BWindowScreen.";
		return;
	}

	mConnected = false;
	mDrawSelector = 0;
	mFlipSelector = 0;
	mFlipTime = (bigtime_t) 0;

	switch (MatchScreenSpace(params.frameBounds, params.pixelFormat))
	{
		case B_8_BIT_640x480:
		case B_15_BIT_640x480:
		case B_16_BIT_640x480:
		case B_32_BIT_640x480:

			mScreen_size.x = 640;
			mScreen_size.y = 480;
			break;

		case B_8_BIT_800x600:
		case B_15_BIT_800x600:
		case B_16_BIT_800x600:
		case B_32_BIT_800x600:

			mScreen_size.x = 800;
			mScreen_size.y = 600;
			break;

		case B_8_BIT_1024x768:
		case B_15_BIT_1024x768:
		case B_16_BIT_1024x768:
		case B_32_BIT_1024x768:

			mScreen_size.x = 1024;
			mScreen_size.y = 768;
			break;

		case B_8_BIT_1152x900:
		case B_15_BIT_1152x900:
		case B_16_BIT_1152x900:
		case B_32_BIT_1152x900:

			mScreen_size.x = 1152;
			mScreen_size.y = 900;
			break;

		case B_8_BIT_1280x1024:
		case B_15_BIT_1280x1024:
		case B_16_BIT_1280x1024:
		case B_32_BIT_1280x1024:

			mScreen_size.x = 1280;
			mScreen_size.y = 1024;
			break;

		case B_8_BIT_1600x1200:
		case B_15_BIT_1600x1200:
		case B_16_BIT_1600x1200:
		case B_32_BIT_1600x1200:

			mScreen_size.x = 1600;
			mScreen_size.y = 1200;
			break;

		default:

			error = B_ERROR;
			errorString = "failure matching a screen space at framework construction.";
			return;
	}

	mClipFrame_size.x = (int32) params.frameBounds.x + 1;
	mClipFrame_size.y = (int32) params.frameBounds.y + 1;

	mClipFrame_upperLeft.x = mScreen_size.x - mClipFrame_size.x >> 1;
	mClipFrame_upperLeft.y = mScreen_size.y - mClipFrame_size.y >> 1;

	mBits = new(std::nothrow) BUFFER_RECORD [params.numRingBuffers];
	if (NULL == mBits)
	{
		error = B_ERROR;
		errorString = "'bad_alloc' at framework construction.";
		return;
	}

	for (uint32 i = 0; i < params.numRingBuffers; i++)
	{
		mBits[i].writeLock = ::kreate_sem(0, NULL, error, errorString);
		mBits[i].unlockAtConnect = true;
	}

	mBScreen = new(std::nothrow) BScreen(this);
	if (NULL == mBScreen)
	{
		error = B_ERROR;
		errorString = "'bad_alloc' at framework construction.";
	}
}

FrmWindowScreen::~FrmWindowScreen()
{
	if (NULL != mBits)
		for (uint32 i = 0; i < mParams.numRingBuffers; i++)
			::delete_sem(mBits[i].writeLock);

	int32 result; 
	thread_id painter = ::find_thread(Framework::__painter_name());

	if (B_NAME_NOT_FOUND != painter)
		::wait_for_thread(painter, &result);

	delete [] mBits;
	delete mBScreen;
}

void FrmWindowScreen::MessageReceived(BMessage *message) 
{
	switch (message->what)
	{
	case 'FLIP':

		if (mConnected)
		{
			if (CanControlFrameBuffer())
			{
				// unwanted things would happen if we rendered pictures at
				// a higher rate than the card can flip. we here make sure
				// not to flip more than 55.5 pictures per second if the
				// card does not support retrace syncing

		      	if (B_OK != mBScreen->WaitForRetrace())
		      	{
					bigtime_t currTime = ::system_time();

					if (currTime < mFlipTime)
					{
						::snooze(mFlipTime - currTime);
						mFlipTime += SANE_FLIP_PERIOD;
					}
					else
						mFlipTime = currTime + SANE_FLIP_PERIOD;
				}		

				MoveDisplayArea(0, mFlipSelector * mScreen_size.y);
			}
			::release_sem(mBits[mFlipSelector].writeLock); 
		}
		else
			mBits[mFlipSelector].unlockAtConnect = true;

		++mFlipSelector %= mParams.numRingBuffers; 
		break;

	default:
		BWindowScreen::MessageReceived(message);
	} 
}

bool FrmWindowScreen::QuitRequested()
{
	Hide(); 
	Sync();
	
	::be_app_messenger.SendMessage(B_QUIT_REQUESTED);

	return BWindowScreen::QuitRequested();
}

void FrmWindowScreen::ScreenConnected(bool connected)
{
	// ScreenConnected() is a regular message hook which runs synchronously to
	// the window looper, hence it is inherently synchronized with the flipping.
	// yet explicit synchronization with the drawing code is needed as drawing
	// takes place right in video memory.

	mDrawLock.Lock();

	if (connected)
		if (CanControlFrameBuffer() && B_OK == SetFrameBuffer(mScreen_size.x, mScreen_size.y * mParams.numRingBuffers))
		{
			graphics_card_info *cardInfo = CardInfo();

			mBitsInfo.bytes_per_row = cardInfo->bytes_per_row;  
			mBitsInfo.bits_per_pixel = cardInfo->bits_per_pixel;

			// initial video buffer cleanup
			::memset(cardInfo->frame_buffer, 0, cardInfo->height * cardInfo->bytes_per_row);

			for (uint32 i = 0; i < mParams.numRingBuffers; i++)
			{
				mBits[i].bits = (uint8 *)cardInfo->frame_buffer + cardInfo->bytes_per_row * mScreen_size.y * i +
									mClipFrame_upperLeft.y * cardInfo->bytes_per_row +
									mClipFrame_upperLeft.x * (cardInfo->bits_per_pixel >> 3); 

				if (mBits[i].unlockAtConnect)
				{
					::release_sem_etc(mBits[i].writeLock, 1, B_DO_NOT_RESCHEDULE);
					mBits[i].unlockAtConnect = false;
				}
			}
		}
		else
		{
			mError = B_ERROR;
			mErrorString = "cannot control frame buffer or number of ring buffers too big for video memory.";

			::be_app_messenger.SendMessage(B_QUIT_REQUESTED);
		}

	mConnected = connected;

	mDrawLock.Unlock();
}

status_t FrmWindowScreen::frame_open(BUFFER_DESC &desc)
{
	status_t res;

	while (B_OK == (res = ::acquire_sem(mBits[mDrawSelector].writeLock)))
	{
		desc.frameHeight	= mClipFrame_size.y;
		desc.frameWidth		= mClipFrame_size.x;
		desc.pixelFormat 	= mParams.pixelFormat;
		desc.rowStride		= mBitsInfo.bytes_per_row;
		desc.pBits			= mBits[mDrawSelector].bits;

		mDrawLock.Lock();

		if (mConnected)
			break;
	
		// acquired buffer is not actually present, try again
		mBits[mDrawSelector].unlockAtConnect = true;
		mDrawLock.Unlock();
	}

	return res; 
}

void FrmWindowScreen::frame_close()
{
	mDrawLock.Unlock();

	// frame drawn, enqueue it for displaying and advance the draw selector
	PostMessage('FLIP');
	++mDrawSelector %= mParams.numRingBuffers; 
}
