#ifndef __RENDCLIPFRUST__
#define __RENDCLIPFRUST__

#include "rendBaseTypes.h"
#include "rendVect.h"

// ClipByW6_ge()	: clips a polygon by one of its vertex coordinates so this coord. is greater or equal to -W
//		- coord,		int32				: index of the coordinate to clip by,		input
//		- poly,			const Vector6 **	: array containing a polygon definition,	input/output
//		- start,		uint32 &			: polygon starting vertex in the array,		input/output
//		- size,			uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,		Vector6 *			: storage buffer for the new vertices,		input/output
//		- nbuff,		uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
inline void ClipByW6_ge(int32 coord,
						const Vector6 *poly[],
						uint32 &start,
						uint32 &size,
						Vector6 buffer[],
						uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		float distCurr = (*poly[currVert])[coord] + (*poly[currVert])[3];
		float distPrev = (*poly[prevVert])[coord] + (*poly[prevVert])[3];

		if (distCurr < 0) // is current vertex outside?
		{
			if (distPrev > 0) // is previous vertex strictly inside?
			{
				Interpolate6(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if (distCurr > 0 &&	// is current vertex strictly inside and
				distPrev < 0)	// is previous vertex outside?
			{
				Interpolate6(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}

// ClipByW6_le()	: clips a polygon by one of its vertex coordinates so this coord. is less or equal to W
//		- coord,		int32				: index of the coordinate to clip,			input
//		- poly,			const Vector6 **	: array containing a polygon definition,	input/output
//		- start,		uint32 &			: polygon starting vertex in the array,		input/output
//		- size,			uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,		Vector6 *			: storage buffer for the new vertices,		input/output
//		- nbuff,		uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
inline void ClipByW6_le(int32 coord,
						const Vector6 *poly[],
						uint32 &start,
						uint32 &size,
						Vector6 buffer[],
						uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		float distCurr = (*poly[currVert])[coord] - (*poly[currVert])[3];
		float distPrev = (*poly[prevVert])[coord] - (*poly[prevVert])[3];

		if (distCurr > 0) // is current vertex outside?
		{
			if (distPrev < 0) // is previous vertex strictly inside?
			{
				Interpolate6(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if (distCurr < 0 &&	// is current vertex strictly inside and
				distPrev > 0)	// is previous vertex outside?
			{
				Interpolate6(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}


// ClipByFrustum6()	: clips a polygon by the view frustum
//		- poly,			const Vector6 **	: array containing a polygon definition,	input/output
//		- start,		uint32 &			: polygon starting vertex in the array,		input/output
//		- size,			uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,		Vector6 *			: storage buffer for the new vertices,		input/output
//		- nbuff,		uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- size would be 0 if clipping discarded all vertices
inline void ClipByFrustum6(const Vector6 *poly[],
						   uint32 &start,
						   uint32 &size,
						   Vector6 buffer[],
						   uint32 &nbuff)
{
	// -W <= X <= W
	ClipByW6_ge(0, poly, start, size, buffer, nbuff);
	ClipByW6_le(0, poly, start, size, buffer, nbuff);

	// -W <= Y <= W
	ClipByW6_ge(1, poly, start, size, buffer, nbuff);
	ClipByW6_le(1, poly, start, size, buffer, nbuff);

	// -W <= Z <= W
	ClipByW6_ge(2, poly, start, size, buffer, nbuff);
	ClipByW6_le(2, poly, start, size, buffer, nbuff);
}

#endif // __RENDCLIPFRUST__
