#include <math.h>
#include "rendSpan.h"
#include "rendClipRect.h"
#include "rendClipFrust.h"
#include "rendVect.h"
#include "rendPrim.h"

// NDCToScreen()	: maps normalized device coordinates to screen space
//		- v,		Vector3 &		: normalized device coordinates,	input/output
// returns
//		nil
// note
//		- mapping is performed accoring to the viewport definition
inline void NDCToScreen(Vector3 &v)
{
	// map the normalized device x and y coordinates to screen coordinates
	v[0] = v[0] * (viewport.screen.right - viewport.screen.left) / 2 +
				  (viewport.screen.right + viewport.screen.left) / 2;

	v[1] = v[1] * (viewport.screen.top - viewport.screen.bottom) / 2 +
				  (viewport.screen.top + viewport.screen.bottom) / 2;

	// map the normalized device z coordinate to the specified depth range
	v[2] = v[2] * (viewport.depth_range_f - viewport.depth_range_n) / 2 +
				  (viewport.depth_range_f + viewport.depth_range_n) / 2;
}

// MapToScreen6()	: maps a vector from clipping space to screen space
//		- vi,		const Vector6 &		: vector in clipping space,	input
//		- vo,		Vector6 &			: vector in screen space,	output
// returns
//		nil
// note
//		- output vector's 4th component is the reciprocal homogeneous W
inline void MapToScreen6(const Vector6 &vi, Vector6 &vo)
{
	// perform perspective division on the clipping-space vertex,
	// normalizing the x, y and z coordinates in the course
	float rw = 1.0f / vi[3];

	vo[0] = vi[0] * rw;
	vo[1] = vi[1] * rw;
	vo[2] = vi[2] * rw;
	vo[3] = rw;
	vo[4] = vi[4] * rw;
	vo[5] = vi[5] * rw;

	// take the normalized device coordinates to screen space
	NDCToScreen(*(Vector3 *)&vo);
}

// Trapezoid6()		: scan-converts a trapezoid polygon
//		- y0_p0,	const Vector6 &		: trapezoid 1st vertex,	input
//		- y0_p1,	const Vector6 &		: trapezoid 2nd vertex,	input
//		- y1_p0,	const Vector6 &		: trapezoid 3rd vertex,	input
//		- y1_p1,	const Vector6 &		: trapezoid 4th vertex,	input
// returns
//		nil
// note
//		- vertices must be in ascending-y-ascending-x order
//		- left edge is used for all height calculations
template < typename SPAN_VERTEX6_T,
		   void (SPAN_ROUTINE_T)(const SPAN_VERTEX6_T &, const SPAN_VERTEX6_T &, int8 *) >

void Trapezoid6(const Vector6 &y0_p0, const Vector6 &y0_p1,
				const Vector6 &y1_p0, const Vector6 &y1_p1)
{
	// snap top end to the next greater- or equal integer;
	// snap bottom end to the previous integer
    const float snap_y0 = (float) ceilf((*(const Vertex2 *)&y0_p0).y);
    const float snap_y1 = (float) ceilf((*(const Vertex2 *)&y1_p0).y) - 1.0f;

    if (snap_y1 < snap_y0)
        return;

    const float err_y0 = snap_y0 - (*(const Vertex2 *)&y0_p0).y;
    const float dy = (*(const Vertex2 *)&y1_p0).y - (*(const Vertex2 *)&y0_p0).y;

	Vector6 dp0;
	dp0[0] = (y1_p0[0] - y0_p0[0]) / dy;
	dp0[2] = (y1_p0[2] - y0_p0[2]) / dy;
	dp0[3] = (y1_p0[3] - y0_p0[3]) / dy;
	dp0[4] = (y1_p0[4] - y0_p0[4]) / dy;
	dp0[5] = (y1_p0[5] - y0_p0[5]) / dy;

	Vector6 dp1;
	dp1[0] = (y1_p1[0] - y0_p1[0]) / dy;
	dp1[2] = (y1_p1[2] - y0_p1[2]) / dy;
	dp1[3] = (y1_p1[3] - y0_p1[3]) / dy;
	dp1[4] = (y1_p1[4] - y0_p1[4]) / dy;
	dp1[5] = (y1_p1[5] - y0_p1[5]) / dy;

    const int trapez_h = (int)(snap_y1 - snap_y0) + 1;
    for (int i = 0; i < trapez_h; i++)
    {
		Vector6 p0i;
		p0i[0] = y0_p0[0] + (err_y0 + (float) i) * dp0[0];
		p0i[2] = y0_p0[2] + (err_y0 + (float) i) * dp0[2];
		p0i[3] = y0_p0[3] + (err_y0 + (float) i) * dp0[3];
		p0i[4] = y0_p0[4] + (err_y0 + (float) i) * dp0[4];
		p0i[5] = y0_p0[5] + (err_y0 + (float) i) * dp0[5];

		Vector6 p1i;
		p1i[0] = y0_p1[0] + (err_y0 + (float) i) * dp1[0];
		p1i[2] = y0_p1[2] + (err_y0 + (float) i) * dp1[2];
		p1i[3] = y0_p1[3] + (err_y0 + (float) i) * dp1[3];
		p1i[4] = y0_p1[4] + (err_y0 + (float) i) * dp1[4];
		p1i[5] = y0_p1[5] + (err_y0 + (float) i) * dp1[5];

		SPAN_ROUTINE_T(*(SPAN_VERTEX6_T *)&p0i, *(SPAN_VERTEX6_T *)&p1i,
						renderState.frame.bits + ((int)snap_y0 + i) * renderState.frame.row_stride);
    }
}

// TriTwiceArea()	: calculates twice the signed area of a 2D triangle,
//					  as the length of the cross-product of two of the sides, taken in the XY plane
//		- p0x,	const float		: X of the triangle's 1st vertex,	input
//		- p0y,	const float		: Y of the triangle's 1st vertex,	input
//		- p1x,	const float		: X of the triangle's 2nd vertex,	input
//		- p1y,	const float		: Y of the triangle's 2nd vertex,	input
//		- p2x,	const float		: X of the triangle's 3rd vertex,	input
//		- p2y,	const float		: Y of the triangle's 3rd vertex,	input
// returns
//		float
//			negative area - counter-clockwise
//			positive area - clockwise
// note
//		- if resulting area's sign will be used as a triangle orienantion determinant,
//		  passed vertices must be in ascending y order
inline float TriTwiceArea(const float p0x, const float p0y,
						  const float p1x, const float p1y,
						  const float p2x, const float p2y)
{
	return (p1x - p0x) * (p2y - p1y) - (p2x - p1x) * (p1y - p0y);
}

// Swap()		: swaps two variables of the same type
//		- a,	<template_type>		: 1st veriable,		input/output
//		- b,	<template_type>		: 2nd veriable,		input/output
// returns
//		nil
template < typename COMMON_T >

inline void Swap(COMMON_T &a, COMMON_T &b)
{
	COMMON_T tmp = a;
	a = b;
	b = tmp;
}

// Tri6()		: scan-converts a triangle, performing culling and port clipping
//		- p0,	const Vector6 &		: triangle 1st vertex,	input
//		- p1,	const Vector6 &		: triangle 2nd vertex,	input
//		- p2,	const Vector6 &		: triangle 3rd vertex,	input
// returns
//		nil
// note
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
template < typename SPAN_VERTEX6_T,
		   void (SPAN_ROUTINE_T)(const SPAN_VERTEX6_T &, const SPAN_VERTEX6_T &, int8 *) >

void Tri6(const Vector6 &p0, const Vector6 &p1, const Vector6 &p2)
{
	// find the ascending-Y order of the passed triangle vertices
	const Vertex2 *v_sorted[3];
	v_sorted[0] = (const Vertex2 *) &p0;
	v_sorted[1] = (const Vertex2 *) &p1;
	v_sorted[2] = (const Vertex2 *) &p2;

	uint32 swapped = 0;

	if ((*v_sorted[0]).y > (*v_sorted[1]).y)
	{
		Swap< const Vertex2 * >(v_sorted[0], v_sorted[1]);
		swapped ^= IEEE754_SIGN_MASK;
	}
	if ((*v_sorted[1]).y > (*v_sorted[2]).y)
	{
		Swap< const Vertex2 * >(v_sorted[1], v_sorted[2]);
		swapped ^= IEEE754_SIGN_MASK;

		// vertex1 & vertex2 swapped - re-check vertex0 & vertex1
		if ((*v_sorted[0]).y > (*v_sorted[1]).y)
		{
			Swap< const Vertex2 * >(v_sorted[0], v_sorted[1]);
			swapped ^= IEEE754_SIGN_MASK;
		}
	}

	// calculate the signed area of the triangle
	float triArea = TriTwiceArea((*v_sorted[0]).x, (*v_sorted[0]).y,
								 (*v_sorted[1]).x, (*v_sorted[1]).y,
								 (*v_sorted[2]).x, (*v_sorted[2]).y);
	(*(uint32*)&triArea) ^= swapped;

	// is zero area?
	if ((*(uint32*)&triArea & ~IEEE754_SIGN_MASK) == 0)
		return;

	// if culling is enabled reject the triangle if its area's sign is the same as the culling sign
	if (renderState.culling)
		if (!((*(uint32*)&triArea ^ renderState.culling) & IEEE754_SIGN_MASK))
			return;

	// clip this triangle by the viewport rectangle
	const uint32 BUFFSIZE = 40;

	const Vector6 *poly[BUFFSIZE];
	Vector6		   buff[BUFFSIZE];

	poly[0] = &p0;
	poly[1] = &p1;
	poly[2] = &p2;

	uint32 start = 0;
	uint32 size = 3;
	uint32 nbuff = 0;

	ClipByRect6(viewport.screen, poly, start, size, buff, nbuff);

	if (size == 0)
		return;

	// calculate the gradients of the vertex parameters (triangle gradients)
	float rcpTriArea = 1.0f / triArea;

	for (uint32 g = 0; g < 4; g++) // iterate the vertex parameters which we need the x-axis and y-axis gradients of
	{
		renderState.trigrad_x[g] = rcpTriArea * TriTwiceArea(p0[2 + g], (*(const Vertex2 *)&p0).y,
															 p1[2 + g], (*(const Vertex2 *)&p1).y,
															 p2[2 + g], (*(const Vertex2 *)&p2).y);

		renderState.trigrad_y[g] = rcpTriArea * TriTwiceArea(p0[2 + g], (*(const Vertex2 *)&p0).x,
															 p1[2 + g], (*(const Vertex2 *)&p1).x,
															 p2[2 + g], (*(const Vertex2 *)&p2).x);
	}

	// find the top-most left and right, and the bottom-most left and right vertices
	int32 vert_ul = start;
	int32 vert_ur = start;
	int32 vert_ll = start;
	int32 vert_lr = start;

	for (uint32 i = start + 1; i < start + size; i++)
	{
		// check for top-most
		if ((*(const Vertex2 *)poly[i]).y < (*(const Vertex2 *)poly[vert_ul]).y) // is this vertex above the top-most?
		{
			vert_ul = i;
			vert_ur = i;
		}
		else
		if ((*(const Vertex2 *)poly[i]).y == (*(const Vertex2 *)poly[vert_ul]).y) // is it at the same y as top-most?
		{
			if ((*(const Vertex2 *)poly[i]).x < (*(const Vertex2 *)poly[vert_ul]).x)
				vert_ul = i;

			if ((*(const Vertex2 *)poly[i]).x > (*(const Vertex2 *)poly[vert_ur]).x)
				vert_ur = i;
		}

		// check for bottom-most
		if ((*(const Vertex2 *)poly[i]).y > (*(const Vertex2 *)poly[vert_ll]).y) // is this vertex below the bottom-most?
		{
			vert_ll = i;
			vert_lr = i;
		}
		else
		if ((*(const Vertex2 *)poly[i]).y == (*(const Vertex2 *)poly[vert_ll]).y) // is it at the same y as bottom-most?
		{
			if ((*(const Vertex2 *)poly[i]).x < (*(const Vertex2 *)poly[vert_ll]).x)
				vert_ll = i;

			if ((*(const Vertex2 *)poly[i]).x > (*(const Vertex2 *)poly[vert_lr]).x)
				vert_lr = i;
		}
	}

	// divide the polygon vertically into non-overlapping trapezoids
	int32 inc_l;

	if (*(uint32 *)&triArea & IEEE754_SIGN_MASK)	// is poly counter-clockwise?
		inc_l = 1;
	else											// poly is clockwise
		inc_l = -1;

	int32 next_l = start + (size + (vert_ul - start) + inc_l) % size;
	int32 next_r = start + (size + (vert_ur - start) - inc_l) % size;

	const Vector6 *top_l = poly[vert_ul];
	const Vector6 *top_r = poly[vert_ur];

	Vector6 *tmp = buff + nbuff;

	while (vert_ul != vert_ll && vert_ur != vert_lr)
	{
		if ((*(const Vertex2 *)poly[next_l]).y == (*(const Vertex2 *)poly[next_r]).y) // do both segments end simultaneously?
		{
			Trapezoid6< SPAN_VERTEX6_T, SPAN_ROUTINE_T >(*top_l, *top_r,
														 *poly[next_l], *poly[next_r]);
			// symetrical advancement
			vert_ul = next_l;
			vert_ur = next_r;

			next_l = start + (size + (next_l - start) + inc_l) % size;
			next_r = start + (size + (next_r - start) - inc_l) % size;

			top_l = poly[vert_ul];
			top_r = poly[vert_ur];
		}
		else
		if ((*(const Vertex2 *)poly[next_l]).y < (*(const Vertex2 *)poly[next_r]).y) // left segment ends first
		{
			IntersectRow6(poly[vert_ur], poly[next_r], (*(const Vertex2 *)poly[next_l]).y, tmp);

			Trapezoid6< SPAN_VERTEX6_T, SPAN_ROUTINE_T >(*top_l, *top_r,
														 *poly[next_l], *tmp);
			// advance left
			vert_ul = next_l;

			next_l = start + (size + (next_l - start) + inc_l) % size;

			top_l = poly[vert_ul];
			top_r = tmp++;
		}
		else // right segment ends first
		{
			IntersectRow6(poly[vert_ul], poly[next_l], (*(const Vertex2 *)poly[next_r]).y, tmp);

			Trapezoid6< SPAN_VERTEX6_T, SPAN_ROUTINE_T >(*top_l, *top_r,
														 *tmp, *poly[next_r]);
			// advance right
			vert_ur = next_r;

			next_r = start + (size + (next_r - start) - inc_l) % size;

			top_l = tmp++;
			top_r = poly[vert_ur];
		}
	}
}

// Triangle6()	: draws a triangle, taking it through a model-projection transform, frustum-clipping it,
//				  culling it and port-clipping it
//		- mat,	const Matrix4 &		: model-projection matrix,	input
//		- p0,	const Vector6 &		: triangle 1st vertex,		input
//		- p1,	const Vector6 &		: triangle 2nd vertex,		input
//		- p2,	const Vector6 &		: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- primitive gets clipped by the view frustum
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
template < typename SPAN_VERTEX6_T,
		   void (SPAN_ROUTINE_T)(const SPAN_VERTEX6_T &, const SPAN_VERTEX6_T &, int8 *) >

void Triangle6(const Matrix4 &mat, const Vector6 &p0, const Vector6 &p1, const Vector6 &p2)
{
	const uint32 BUFFSIZE = 40;

	Vector6 p_clip[BUFFSIZE];

	CopyTransform6(mat, p0, p_clip[0]);
	CopyTransform6(mat, p1, p_clip[1]);
	CopyTransform6(mat, p2, p_clip[2]);

	const Vector6 *poly[BUFFSIZE];

	poly[0] = p_clip + 0;
	poly[1] = p_clip + 1;
	poly[2] = p_clip + 2;

	uint32 start = 0;
	uint32 size = 3;
	uint32 nclip = 3;

	ClipByFrustum6(poly, start, size, p_clip, nclip);

	if (size == 0)
		return;

	Vector6 p_scr[BUFFSIZE];

	for (uint32 i = 0; i < size; i++)
		MapToScreen6(*poly[start + i], p_scr[i]);

	// rasterize the resulting polygon as a triangle fan
	for (uint32 i = 0; i < size - 2; i++)
		Tri6< SPAN_VERTEX6_T, SPAN_ROUTINE_T >(p_scr[0], p_scr[i + 1], p_scr[i + 2]);
}

// PlotTriangle6()	: transforms and draws a triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- mat,	const Matrix4 &		: model-projection matrix,	input
//		- p0,	const Vector6 &		: triangle 1st vertex,		input
//		- p1,	const Vector6 &		: triangle 2nd vertex,		input
//		- p2,	const Vector6 &		: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- primitive gets clipped by the view frustum
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle6(const Matrix4 &mat, const Vector6 &p0, const Vector6 &p1, const Vector6 &p2)
{
	switch((renderState.sampling << 24) | (renderState.frame.pix_format << 16) | (renderState.vert_format << 8) | renderState.color_op)
	{
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)	   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:

		Triangle6< Vertex4UV, Span4UV_any_nop_any >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_cpy_argb1555 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_cpy_argb8888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_cpy_rgb565 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_cpy_rgb888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_linear_cpy_argb1555 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_linear_cpy_argb8888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_linear_cpy_rgb565 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_linear_cpy_rgb888 >(mat, p0, p1, p2);
		break;
		
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_nearest_cpy_argb8888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_nearest_cpy_argb1555 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_nearest_cpy_rgb888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_nearest_cpy_rgb565 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_linear_cpy_argb8888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_linear_cpy_argb1555 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_linear_cpy_rgb888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle6< Vertex4UV, Span4UV_nearest_linear_cpy_rgb565 >(mat, p0, p1, p2);
		break;
	}
}

// PlotTriangle6()	: draws a screen-space triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- p0,	const Vector6 &		: triangle 1st vertex,		input
//		- p1,	const Vector6 &		: triangle 2nd vertex,		input
//		- p2,	const Vector6 &		: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- vectors' 3rd element (Z) must be in the range [0, 1], and 4th element (RHW) must be non-zero
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle6(const Vector6 &p0, const Vector6 &p1, const Vector6 &p2)
{
	switch((renderState.sampling << 24) | (renderState.frame.pix_format << 16) | (renderState.vert_format << 8) | renderState.color_op)
	{
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)	   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:

		Tri6< Vertex4UV, Span4UV_any_nop_any >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_cpy_argb1555 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_cpy_argb8888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_cpy_rgb565 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_cpy_rgb888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_linear_cpy_argb1555 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_linear_cpy_argb8888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_linear_cpy_rgb565 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_linear_cpy_rgb888 >(p0, p1, p2);
		break;
		
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_nearest_cpy_argb8888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_nearest_cpy_argb1555 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_nearest_cpy_rgb888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_nearest_cpy_rgb565 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_linear_cpy_argb8888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_linear_cpy_argb1555 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_linear_cpy_rgb888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri6< Vertex4UV, Span4UV_nearest_linear_cpy_rgb565 >(p0, p1, p2);
		break;
	}
}
