#ifndef __RENDVECT__
#define __RENDVECT__

#include <math.h>
#include <memory.h>
#include "rendBaseTypes.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
// remark on conventions
//
// vectors are taken as column-vectors and matrices as column-major, thus
// matrix transformations follow operator-leftism, and are written out as
// Tn * Tn-1 * .. * T1 * T0 * v
////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////
// prototypes
////////////////////////////////////////////////////////////////////////////////////////////////////

void Rotate(Matrix4 &, float, float, float, float);
void Translate(Matrix4 &, float, float, float);
void Scale(Matrix4 &, float, float, float);

void Frustum(Matrix4 &, float, float, float, float, float, float);
void Ortho(Matrix4 &, float, float, float, float, float, float);

void CopyInvertMat(const Matrix4 &, Matrix4 &);

////////////////////////////////////////////////////////////////////////////////////////////////////
// templates & inlines
////////////////////////////////////////////////////////////////////////////////////////////////////

inline void MulVect4(Vector4 &v, float scalar)
{
	v[0] *= scalar;
	v[1] *= scalar;
	v[2] *= scalar;
	v[3] *= scalar;
}

template < int STRIDE_V0_T, int STRIDE_V1_T >
float DotVect4(const float *v0, const float *v1)
{
	return v0[STRIDE_V0_T * 0] * v1[STRIDE_V1_T * 0] +
		   v0[STRIDE_V0_T * 1] * v1[STRIDE_V1_T * 1] +
		   v0[STRIDE_V0_T * 2] * v1[STRIDE_V1_T * 2] +
		   v0[STRIDE_V0_T * 3] * v1[STRIDE_V1_T * 3];
}

inline float DotVect4(const Vector4 &v0, const Vector4 &v1)
{
    return DotVect4< 1, 1 >(v0, v1);
}

inline float LenSqrVect4(const Vector4 &v)
{
	return DotVect4(v, v);
}

inline float LenVect4(const Vector4 &v)
{
	return sqrtf(LenSqrVect4(v));
}

inline void LoadMat(Matrix4 &mat,
					float a00, float a01, float a02, float a03,
					float a10, float a11, float a12, float a13,
					float a20, float a21, float a22, float a23,
					float a30, float a31, float a32, float a33)
{
	mat[0][0] = a00;
	mat[0][1] = a01;
	mat[0][2] = a02;
	mat[0][3] = a03;

	mat[1][0] = a10;
	mat[1][1] = a11;
	mat[1][2] = a12;
	mat[1][3] = a13;

	mat[2][0] = a20;
	mat[2][1] = a21;
	mat[2][2] = a22;
	mat[2][3] = a23;

	mat[3][0] = a30;
	mat[3][1] = a31;
	mat[3][2] = a32;
	mat[3][3] = a33;
}

inline void LoadMat(Matrix4 &dst, const Matrix4 &src)
{
	memcpy(&dst, &src, sizeof(Matrix4));
}

inline void MulMat4(const Matrix4 &matA, Matrix4 &matB)
{
	Vector4 colB[4];

	colB[0][0] = matB[0][0];
	colB[0][1] = matB[1][0];
	colB[0][2] = matB[2][0];
	colB[0][3] = matB[3][0];

	colB[1][0] = matB[0][1];
	colB[1][1] = matB[1][1];
	colB[1][2] = matB[2][1];
	colB[1][3] = matB[3][1];

	colB[2][0] = matB[0][2];
	colB[2][1] = matB[1][2];
	colB[2][2] = matB[2][2];
	colB[2][3] = matB[3][2];

	colB[3][0] = matB[0][3];
	colB[3][1] = matB[1][3];
	colB[3][2] = matB[2][3];
	colB[3][3] = matB[3][3];

	matB[0][0] = DotVect4(matA[0], colB[0]);
	matB[0][1] = DotVect4(matA[0], colB[1]);
	matB[0][2] = DotVect4(matA[0], colB[2]);
	matB[0][3] = DotVect4(matA[0], colB[3]);

	matB[1][0] = DotVect4(matA[1], colB[0]);
	matB[1][1] = DotVect4(matA[1], colB[1]);
	matB[1][2] = DotVect4(matA[1], colB[2]);
	matB[1][3] = DotVect4(matA[1], colB[3]);

	matB[2][0] = DotVect4(matA[2], colB[0]);
	matB[2][1] = DotVect4(matA[2], colB[1]);
	matB[2][2] = DotVect4(matA[2], colB[2]);
	matB[2][3] = DotVect4(matA[2], colB[3]);

	matB[3][0] = DotVect4(matA[3], colB[0]);
	matB[3][1] = DotVect4(matA[3], colB[1]);
	matB[3][2] = DotVect4(matA[3], colB[2]);
	matB[3][3] = DotVect4(matA[3], colB[3]);
}

inline void MulMat4(const Matrix4 &matA, const Matrix4 &matB, Matrix4 &matC)
{
	matC[0][0] = DotVect4< 1, 4 >(matA[0], matB[0] + 0);
	matC[0][1] = DotVect4< 1, 4 >(matA[0], matB[0] + 1);
	matC[0][2] = DotVect4< 1, 4 >(matA[0], matB[0] + 2);
	matC[0][3] = DotVect4< 1, 4 >(matA[0], matB[0] + 3);

	matC[1][0] = DotVect4< 1, 4 >(matA[1], matB[0] + 0);
	matC[1][1] = DotVect4< 1, 4 >(matA[1], matB[0] + 1);
	matC[1][2] = DotVect4< 1, 4 >(matA[1], matB[0] + 2);
	matC[1][3] = DotVect4< 1, 4 >(matA[1], matB[0] + 3);

	matC[2][0] = DotVect4< 1, 4 >(matA[2], matB[0] + 0);
	matC[2][1] = DotVect4< 1, 4 >(matA[2], matB[0] + 1);
	matC[2][2] = DotVect4< 1, 4 >(matA[2], matB[0] + 2);
	matC[2][3] = DotVect4< 1, 4 >(matA[2], matB[0] + 3);

	matC[3][0] = DotVect4< 1, 4 >(matA[3], matB[0] + 0);
	matC[3][1] = DotVect4< 1, 4 >(matA[3], matB[0] + 1);
	matC[3][2] = DotVect4< 1, 4 >(matA[3], matB[0] + 2);
	matC[3][3] = DotVect4< 1, 4 >(matA[3], matB[0] + 3);
}

inline void CopyTransform6(const Matrix4 &mat, const Vector6 &vi, Vector6 &vo)
{
	vo[0] = DotVect4(mat[0], *(const Vector4 *)&vi);
	vo[1] = DotVect4(mat[1], *(const Vector4 *)&vi);
	vo[2] = DotVect4(mat[2], *(const Vector4 *)&vi);
	vo[3] = DotVect4(mat[3], *(const Vector4 *)&vi);
	vo[4] = vi[4];
	vo[5] = vi[5];
}

inline void Interpolate6(const Vector6 &p0,
						 const Vector6 &p1,
						 float dist0,
						 float dist1,
						 Vector6 &p_interpol)
{
	// take the absolute value of the distances
	*(uint32 *)&dist0 &= ~IEEE754_SIGN_MASK;
	*(uint32 *)&dist1 &= ~IEEE754_SIGN_MASK;
	
	float mass0 = dist1 / (dist0 + dist1);
	float mass1 = dist0 / (dist0 + dist1);

	p_interpol[0] = p0[0] * mass0 + p1[0] * mass1;
	p_interpol[1] = p0[1] * mass0 + p1[1] * mass1;
	p_interpol[2] = p0[2] * mass0 + p1[2] * mass1;
	p_interpol[3] = p0[3] * mass0 + p1[3] * mass1;
	p_interpol[4] = p0[4] * mass0 + p1[4] * mass1;
	p_interpol[5] = p0[5] * mass0 + p1[5] * mass1;
}

// ClipByPlane6()	: clips a polygon by a plane, halfspace of negative distance is 'inside'
//		- plane,		const Vector4 &		: halfspaces border,						input
//		- poly,			const Vector6 **	: array containing a polygon definition,	input/output
//		- start,		uint32 &			: polygon starting vertex in the array,		input/output
//		- size,			uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,		Vector6 *			: storage buffer for the new vertices,		input/output
//		- nbuff,		uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- plane vector defined as (n.x, n.y, n.z, -d), where |n| = 1;
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
inline void ClipByPlane6(const Vector4 &plane,
						 const Vector6 *poly[],
						 uint32 &start,
						 uint32 &size,
						 Vector6 buffer[],
						 uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		float distCurr = DotVect4((*(const Vector4 *)poly[currVert]), plane);
		float distPrev = DotVect4((*(const Vector4 *)poly[prevVert]), plane);

		if (distCurr > 0) // is current vertex outside?
		{
			if (distPrev < 0) // is previous vertex strictly inside?
			{
				Interpolate6(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if (distCurr < 0 &&	// is current vertex strictly inside and
				distPrev > 0)	// is previous vertex outside?
			{
				Interpolate6(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}

#endif // __RENDVECT__
